#-*- coding: utf-8 -*-

from ctypes import POINTER, c_void_p, c_uint, c_char, c_float, Structure, c_char_p, c_double, c_ubyte, c_size_t, c_uint32


class Vector2D(Structure):
    """
    See 'vector2.h' for details.
    """


    _fields_ = [
            ("x", c_float),("y", c_float),
        ]

class Matrix3x3(Structure):
    """
    See 'matrix3x3.h' for details.
    """


    _fields_ = [
            ("a1", c_float),("a2", c_float),("a3", c_float),
            ("b1", c_float),("b2", c_float),("b3", c_float),
            ("c1", c_float),("c2", c_float),("c3", c_float),
        ]

class Texel(Structure):
    """
    See 'texture.h' for details.
    """

    _fields_ = [
            ("b", c_ubyte),("g", c_ubyte),("r", c_ubyte),("a", c_ubyte),
        ]

class Color4D(Structure):
    """
    See 'color4.h' for details.
    """


    _fields_ = [
            #  Red, green, blue and alpha color values
            ("r", c_float),("g", c_float),("b", c_float),("a", c_float),
        ]

class Plane(Structure):
    """
    See 'types.h' for details.
    """

    _fields_ = [
            #  Plane equation
            ("a", c_float),("b", c_float),("c", c_float),("d", c_float),
        ]

class Color3D(Structure):
    """
    See 'types.h' for details.
    """

    _fields_ = [
            #  Red, green and blue color values
            ("r", c_float),("g", c_float),("b", c_float),
        ]

class String(Structure):
    """
    See 'types.h' for details.
    """

    MAXLEN = 1024

    _fields_ = [
            # Binary length of the string excluding the terminal 0. This is NOT the
            #  logical length of strings containing UTF-8 multibyte sequences! It's
            #  the number of bytes from the beginning of the string to its end.
            ("length", c_uint32),

            # String buffer. Size limit is MAXLEN
            ("data", c_char*MAXLEN),
        ]

class MaterialPropertyString(Structure):
    """
    See 'MaterialSystem.cpp' for details.

    The size of length is truncated to 4 bytes on 64-bit platforms when used as a
    material property (see MaterialSystem.cpp aiMaterial::AddProperty() for details).
    """

    MAXLEN = 1024

    _fields_ = [
            # Binary length of the string excluding the terminal 0. This is NOT the
            #  logical length of strings containing UTF-8 multibyte sequences! It's
            #  the number of bytes from the beginning of the string to its end.
            ("length", c_uint32),

            # String buffer. Size limit is MAXLEN
            ("data", c_char*MAXLEN),
        ]

class MemoryInfo(Structure):
    """
    See 'types.h' for details.
    """

    _fields_ = [
            # Storage allocated for texture data
            ("textures", c_uint),

            # Storage allocated for material data
            ("materials", c_uint),

            # Storage allocated for mesh data
            ("meshes", c_uint),

            # Storage allocated for node data
            ("nodes", c_uint),

            # Storage allocated for animation data
            ("animations", c_uint),

            # Storage allocated for camera data
            ("cameras", c_uint),

            # Storage allocated for light data
            ("lights", c_uint),

            # Total storage allocated for the full import.
            ("total", c_uint),
        ]

class Quaternion(Structure):
    """
    See 'quaternion.h' for details.
    """


    _fields_ = [
            #  w,x,y,z components of the quaternion
            ("w", c_float),("x", c_float),("y", c_float),("z", c_float),
        ]

class Face(Structure):
    """
    See 'mesh.h' for details.
    """

    _fields_ = [
            #  Number of indices defining this face.
            #  The maximum value for this member is
            #AI_MAX_FACE_INDICES.
            ("mNumIndices", c_uint),

            #  Pointer to the indices array. Size of the array is given in numIndices.
            ("mIndices", POINTER(c_uint)),
        ]

class VertexWeight(Structure):
    """
    See 'mesh.h' for details.
    """

    _fields_ = [
            #  Index of the vertex which is influenced by the bone.
            ("mVertexId", c_uint),

            #  The strength of the influence in the range (0...1).
            #  The influence from all bones at one vertex amounts to 1.
            ("mWeight", c_float),
        ]

class Matrix4x4(Structure):
    """
    See 'matrix4x4.h' for details.
    """


    _fields_ = [
            ("a1", c_float),("a2", c_float),("a3", c_float),("a4", c_float),
            ("b1", c_float),("b2", c_float),("b3", c_float),("b4", c_float),
            ("c1", c_float),("c2", c_float),("c3", c_float),("c4", c_float),
            ("d1", c_float),("d2", c_float),("d3", c_float),("d4", c_float),
        ]

class Vector3D(Structure):
    """
    See 'vector3.h' for details.
    """


    _fields_ = [
            ("x", c_float),("y", c_float),("z", c_float),
        ]

class MeshKey(Structure):
    """
    See 'anim.h' for details.
    """

    _fields_ = [
            # The time of this key
            ("mTime", c_double),

            # Index into the aiMesh::mAnimMeshes array of the
            #  mesh corresponding to the
            #aiMeshAnim hosting this
            #  key frame. The referenced anim mesh is evaluated
            #  according to the rules defined in the docs for
            #aiAnimMesh.
            ("mValue", c_uint),
        ]

class MetadataEntry(Structure):
    """
    See 'metadata.h' for details
    """
    AI_BOOL       = 0
    AI_INT32      = 1
    AI_UINT64     = 2
    AI_FLOAT      = 3
    AI_DOUBLE     = 4
    AI_AISTRING   = 5
    AI_AIVECTOR3D = 6
    AI_META_MAX   = 7
    _fields_ = [
            # The type field uniquely identifies the underlying type of the data field
            ("mType", c_uint),
            ("mData", c_void_p),
    ]

class Metadata(Structure):
    """
    See 'metadata.h' for details
    """
    _fields_ = [
            # Length of the mKeys and mValues arrays, respectively
            ("mNumProperties", c_uint),

            # Arrays of keys, may not be NULL. Entries in this array may not be NULL
            # as well.
            ("mKeys", POINTER(String)),

            # Arrays of values, may not be NULL. Entries in this array may be NULL
            # if the corresponding property key has no assigned value.
            ("mValues", POINTER(MetadataEntry)),
    ]

class Node(Structure):
    """
    See 'scene.h' for details.
    """


Node._fields_ = [
            # The name of the node.
            # The name might be empty (length of zero) but all nodes which
            # need to be accessed afterwards by bones or anims are usually named.
            # Multiple nodes may have the same name, but nodes which are accessed
            # by bones (see
            #aiBone and
            #aiMesh::mBones) *must* be unique.
            # Cameras and lights are assigned to a specific node name - if there
            # are multiple nodes with this name, they're assigned to each of them.
            # <br>
            # There are no limitations regarding the characters contained in
            # this text. You should be able to handle stuff like whitespace, tabs,
            # linefeeds, quotation marks, ampersands, ... .
            ("mName", String),

            # The transformation relative to the node's parent.
            ("mTransformation", Matrix4x4),

            # Parent node. NULL if this node is the root node.
            ("mParent", POINTER(Node)),

            # The number of child nodes of this node.
            ("mNumChildren", c_uint),

            # The child nodes of this node. NULL if mNumChildren is 0.
            ("mChildren", POINTER(POINTER(Node))),

            # The number of meshes of this node.
            ("mNumMeshes", c_uint),

            # The meshes of this node. Each entry is an index into the mesh
            ("mMeshes", POINTER(c_uint)),

            # Metadata associated with this node or NULL if there is no metadata.
            # Whether any metadata is generated depends on the source file format.
            ("mMetadata", POINTER(Metadata)),
        ]

class Light(Structure):
    """
    See 'light.h' for details.
    """


    _fields_ = [
            # The name of the light source.
            #  There must be a node in the scenegraph with the same name.
            #  This node specifies the position of the light in the scene
            #  hierarchy and can be animated.
            ("mName", String),

            # The type of the light source.
            # aiLightSource_UNDEFINED is not a valid value for this member.
            ("mType", c_uint),

            # Position of the light source in space. Relative to the
            #  transformation of the node corresponding to the light.
            #  The position is undefined for directional lights.
            ("mPosition", Vector3D),

            # Direction of the light source in space. Relative to the
            #  transformation of the node corresponding to the light.
            #  The direction is undefined for point lights. The vector
            #  may be normalized, but it needn't.
            ("mDirection", Vector3D),

            # Up direction of the light source in space. Relative to the
            #  transformation of the node corresponding to the light.
            #
            # The direction is undefined for point lights. The vector
            #  may be normalized, but it needn't.
            ("mUp", Vector3D),

            # Constant light attenuation factor.
            #  The intensity of the light source at a given distance 'd' from
            #  the light's position is
            #  @code
            #  Atten = 1/( att0 + att1
            # d + att2
            # d*d)
            #  @endcode
            #  This member corresponds to the att0 variable in the equation.
            #  Naturally undefined for directional lights.
            ("mAttenuationConstant", c_float),

            # Linear light attenuation factor.
            #  The intensity of the light source at a given distance 'd' from
            #  the light's position is
            #  @code
            #  Atten = 1/( att0 + att1
            # d + att2
            # d*d)
            #  @endcode
            #  This member corresponds to the att1 variable in the equation.
            #  Naturally undefined for directional lights.
            ("mAttenuationLinear", c_float),

            # Quadratic light attenuation factor.
            #  The intensity of the light source at a given distance 'd' from
            #  the light's position is
            #  @code
            #  Atten = 1/( att0 + att1
            # d + att2
            # d*d)
            #  @endcode
            #  This member corresponds to the att2 variable in the equation.
            #  Naturally undefined for directional lights.
            ("mAttenuationQuadratic", c_float),

            # Diffuse color of the light source
            #  The diffuse light color is multiplied with the diffuse
            #  material color to obtain the final color that contributes
            #  to the diffuse shading term.
            ("mColorDiffuse", Color3D),

            # Specular color of the light source
            #  The specular light color is multiplied with the specular
            #  material color to obtain the final color that contributes
            #  to the specular shading term.
            ("mColorSpecular", Color3D),

            # Ambient color of the light source
            #  The ambient light color is multiplied with the ambient
            #  material color to obtain the final color that contributes
            #  to the ambient shading term. Most renderers will ignore
            #  this value it, is just a remaining of the fixed-function pipeline
            #  that is still supported by quite many file formats.
            ("mColorAmbient", Color3D),

            # Inner angle of a spot light's light cone.
            #  The spot light has maximum influence on objects inside this
            #  angle. The angle is given in radians. It is 2PI for point
            #  lights and undefined for directional lights.
            ("mAngleInnerCone", c_float),

            # Outer angle of a spot light's light cone.
            #  The spot light does not affect objects outside this angle.
            #  The angle is given in radians. It is 2PI for point lights and
            #  undefined for directional lights. The outer angle must be
            #  greater than or equal to the inner angle.
            #  It is assumed that the application uses a smooth
            #  interpolation between the inner and the outer cone of the
            #  spot light.
            ("mAngleOuterCone", c_float),

            # Size of area light source.
            ("mSize", Vector2D),
        ]

class Texture(Structure):
    """
    See 'texture.h' for details.
    """


    _fields_ = [
            # Width of the texture, in pixels
            # If mHeight is zero the texture is compressed in a format
            # like JPEG. In this case mWidth specifies the size of the
            # memory area pcData is pointing to, in bytes.
            ("mWidth", c_uint),

            # Height of the texture, in pixels
            # If this value is zero, pcData points to an compressed texture
            # in any format (e.g. JPEG).
            ("mHeight", c_uint),

            # A hint from the loader to make it easier for applications
            # to determine the type of embedded textures.
            #
            # If mHeight != 0 this member is show how data is packed. Hint will consist of
            # two parts: channel order and channel bitness (count of the bits for every
            # color channel). For simple parsing by the viewer it's better to not omit
            # absent color channel and just use 0 for bitness. For example:
            # 1. Image contain RGBA and 8 bit per channel, achFormatHint == "rgba8888";
            # 2. Image contain ARGB and 8 bit per channel, achFormatHint == "argb8888";
            # 3. Image contain RGB and 5 bit for R and B channels and 6 bit for G channel,
            #    achFormatHint == "rgba5650";
            # 4. One color image with B channel and 1 bit for it, achFormatHint == "rgba0010";
            # If mHeight == 0 then achFormatHint is set set to '\\0\\0\\0\\0' if the loader has no additional
            # information about the texture file format used OR the
            # file extension of the format without a trailing dot. If there
            # are multiple file extensions for a format, the shortest
            # extension is chosen (JPEG maps to 'jpg', not to 'jpeg').
            # E.g. 'dds\\0', 'pcx\\0', 'jpg\\0'.  All characters are lower-case.
            # The fourth character will always be '\\0'.
            ("achFormatHint", c_char*9),

            # Data of the texture.
            # Points to an array of mWidth
            # mHeight aiTexel's.
            # The format of the texture data is always ARGB8888 to
            # make the implementation for user of the library as easy
            # as possible. If mHeight = 0 this is a pointer to a memory
            # buffer of size mWidth containing the compressed texture
            # data. Good luck, have fun!
            ("pcData", POINTER(Texel)),

            # Texture original filename
            # Used to get the texture reference
            ("mFilename", String),
        ]

class Ray(Structure):
    """
    See 'types.h' for details.
    """

    _fields_ = [
            #  Position and direction of the ray
            ("pos", Vector3D),("dir", Vector3D),
        ]

class UVTransform(Structure):
    """
    See 'material.h' for details.
    """

    _fields_ = [
            # Translation on the u and v axes.
            #  The default value is (0|0).
            ("mTranslation", Vector2D),

            # Scaling on the u and v axes.
            #  The default value is (1|1).
            ("mScaling", Vector2D),

            # Rotation - in counter-clockwise direction.
            #  The rotation angle is specified in radians. The
            #  rotation center is 0.5f|0.5f. The default value
            #  0.f.
            ("mRotation", c_float),
        ]

class MaterialProperty(Structure):
    """
    See 'material.h' for details.
    """

    _fields_ = [
            # Specifies the name of the property (key)
            #  Keys are generally case insensitive.
            ("mKey", String),

            # Textures: Specifies their exact usage semantic.
            # For non-texture properties, this member is always 0
            # (or, better-said,
            #aiTextureType_NONE).
            ("mSemantic", c_uint),

            # Textures: Specifies the index of the texture.
            #  For non-texture properties, this member is always 0.
            ("mIndex", c_uint),

            # Size of the buffer mData is pointing to, in bytes.
            #  This value may not be 0.
            ("mDataLength", c_uint),

            # Type information for the property.
            # Defines the data layout inside the data buffer. This is used
            # by the library internally to perform debug checks and to
            # utilize proper type conversions.
            # (It's probably a hacky solution, but it works.)
            ("mType", c_uint),

            # Binary buffer to hold the property's value.
            # The size of the buffer is always mDataLength.
            ("mData", POINTER(c_char)),
        ]

class Material(Structure):
    """
    See 'material.h' for details.
    """

    _fields_ = [
            # List of all material properties loaded.
            ("mProperties", POINTER(POINTER(MaterialProperty))),

            # Number of properties in the data base
            ("mNumProperties", c_uint),

            # Storage allocated
            ("mNumAllocated", c_uint),
        ]

class Bone(Structure):
    """
    See 'mesh.h' for details.
    """

    _fields_ = [
            #  The name of the bone.
            ("mName", String),

            #  The number of vertices affected by this bone
            #  The maximum value for this member is
            #AI_MAX_BONE_WEIGHTS.
            ("mNumWeights", c_uint),

            #  The vertices affected by this bone
            ("mWeights", POINTER(VertexWeight)),

            #  Matrix that transforms from mesh space to bone space in bind pose
            ("mOffsetMatrix", Matrix4x4),
        ]


class AnimMesh(Structure):
    """
    See 'mesh.h' for details.
    """

    AI_MAX_NUMBER_OF_TEXTURECOORDS = 0x8
    AI_MAX_NUMBER_OF_COLOR_SETS = 0x8

    _fields_ = [
            #  Anim Mesh name
            ("mName", String),

            # Replacement for aiMesh::mVertices. If this array is non-NULL,
            # it *must* contain mNumVertices entries. The corresponding
            # array in the host mesh must be non-NULL as well - animation
            # meshes may neither add or nor remove vertex components (if
            # a replacement array is NULL and the corresponding source
            # array is not, the source data is taken instead)
            ("mVertices", POINTER(Vector3D)),

            # Replacement for aiMesh::mNormals.
            ("mNormals", POINTER(Vector3D)),

            # Replacement for aiMesh::mTangents.
            ("mTangents", POINTER(Vector3D)),

            # Replacement for aiMesh::mBitangents.
            ("mBitangents", POINTER(Vector3D)),

            # Replacement for aiMesh::mColors
            ("mColors", POINTER(Color4D) * AI_MAX_NUMBER_OF_COLOR_SETS),

            # Replacement for aiMesh::mTextureCoords
            ("mTextureCoords", POINTER(Vector3D) * AI_MAX_NUMBER_OF_TEXTURECOORDS),

            # The number of vertices in the aiAnimMesh, and thus the length of all
            # the member arrays.
            #
            # This has always the same value as the mNumVertices property in the
            # corresponding aiMesh. It is duplicated here merely to make the length
            # of the member arrays accessible even if the aiMesh is not known, e.g.
            # from language bindings.
            ("mNumVertices", c_uint),

            # Weight of the AnimMesh.
            ("mWeight", c_float),
        ]


class Mesh(Structure):
    """
    See 'mesh.h' for details.
    """

    AI_MAX_FACE_INDICES = 0x7fff
    AI_MAX_BONE_WEIGHTS = 0x7fffffff
    AI_MAX_VERTICES = 0x7fffffff
    AI_MAX_FACES = 0x7fffffff
    AI_MAX_NUMBER_OF_COLOR_SETS = 0x8
    AI_MAX_NUMBER_OF_TEXTURECOORDS = 0x8

    _fields_ = [ # Bitwise combination of the members of the
            #aiPrimitiveType enum.
            # This specifies which types of primitives are present in the mesh.
            # The "SortByPrimitiveType"-Step can be used to make sure the
            # output meshes consist of one primitive type each.
            ("mPrimitiveTypes", c_uint),

            # The number of vertices in this mesh.
            # This is also the size of all of the per-vertex data arrays.
            # The maximum value for this member is
            #AI_MAX_VERTICES.
            ("mNumVertices", c_uint),

            # The number of primitives (triangles, polygons, lines) in this  mesh.
            # This is also the size of the mFaces array.
            # The maximum value for this member is
            #AI_MAX_FACES.
            ("mNumFaces", c_uint),

            # Vertex positions.
            # This array is always present in a mesh. The array is
            # mNumVertices in size.
            ("mVertices", POINTER(Vector3D)),

            # Vertex normals.
            # The array contains normalized vectors, NULL if not present.
            # The array is mNumVertices in size. Normals are undefined for
            # point and line primitives. A mesh consisting of points and
            # lines only may not have normal vectors. Meshes with mixed
            # primitive types (i.e. lines and triangles) may have normals,
            # but the normals for vertices that are only referenced by
            # point or line primitives are undefined and set to QNaN (WARN:
            # qNaN compares to inequal to *everything*, even to qNaN itself.
            # Using code like this to check whether a field is qnan is:
            # @code
            #define IS_QNAN(f) (f != f)
            # @endcode
            # still dangerous because even 1.f == 1.f could evaluate to false! (
            # remember the subtleties of IEEE754 artithmetics). Use stuff like
            # @c fpclassify instead.
            # @note Normal vectors computed by Assimp are always unit-length.
            # However, this needn't apply for normals that have been taken
            #   directly from the model file.
            ("mNormals", POINTER(Vector3D)),

            # Vertex tangents.
            # The tangent of a vertex points in the direction of the positive
            # X texture axis. The array contains normalized vectors, NULL if
            # not present. The array is mNumVertices in size. A mesh consisting
            # of points and lines only may not have normal vectors. Meshes with
            # mixed primitive types (i.e. lines and triangles) may have
            # normals, but the normals for vertices that are only referenced by
            # point or line primitives are undefined and set to qNaN.  See
            # the
            #mNormals member for a detailed discussion of qNaNs.
            # @note If the mesh contains tangents, it automatically also
            # contains bitangents (the bitangent is just the cross product of
            # tangent and normal vectors).
            ("mTangents", POINTER(Vector3D)),

            # Vertex bitangents.
            # The bitangent of a vertex points in the direction of the positive
            # Y texture axis. The array contains normalized vectors, NULL if not
            # present. The array is mNumVertices in size.
            # @note If the mesh contains tangents, it automatically also contains
            # bitangents.
            ("mBitangents", POINTER(Vector3D)),

            # Vertex color sets.
            # A mesh may contain 0 to
            #AI_MAX_NUMBER_OF_COLOR_SETS vertex
            # colors per vertex. NULL if not present. Each array is
            # mNumVertices in size if present.
            ("mColors", POINTER(Color4D)*AI_MAX_NUMBER_OF_COLOR_SETS),

            # Vertex texture coords, also known as UV channels.
            # A mesh may contain 0 to AI_MAX_NUMBER_OF_TEXTURECOORDS per
            # vertex. NULL if not present. The array is mNumVertices in size.
            ("mTextureCoords", POINTER(Vector3D)*AI_MAX_NUMBER_OF_TEXTURECOORDS),

            # Specifies the number of components for a given UV channel.
            # Up to three channels are supported (UVW, for accessing volume
            # or cube maps). If the value is 2 for a given channel n, the
            # component p.z of mTextureCoords[n][p] is set to 0.0f.
            # If the value is 1 for a given channel, p.y is set to 0.0f, too.
            # @note 4D coords are not supported
            ("mNumUVComponents", c_uint*AI_MAX_NUMBER_OF_TEXTURECOORDS),

            # The faces the mesh is constructed from.
            # Each face refers to a number of vertices by their indices.
            # This array is always present in a mesh, its size is given
            # in mNumFaces. If the
            #AI_SCENE_FLAGS_NON_VERBOSE_FORMAT
            # is NOT set each face references an unique set of vertices.
            ("mFaces", POINTER(Face)),

            # The number of bones this mesh contains.
            # Can be 0, in which case the mBones array is NULL.
            ("mNumBones", c_uint),

            # The bones of this mesh.
            # A bone consists of a name by which it can be found in the
            # frame hierarchy and a set of vertex weights.
            ("mBones", POINTER(POINTER(Bone))),

            # The material used by this mesh.
            # A mesh does use only a single material. If an imported model uses
            # multiple materials, the import splits up the mesh. Use this value
            # as index into the scene's material list.
            ("mMaterialIndex", c_uint),

            # Name of the mesh. Meshes can be named, but this is not a
            #  requirement and leaving this field empty is totally fine.
            #  There are mainly three uses for mesh names:
            #   - some formats name nodes and meshes independently.
            #   - importers tend to split meshes up to meet the
            #      one-material-per-mesh requirement. Assigning
            #      the same (dummy) name to each of the result meshes
            #      aids the caller at recovering the original mesh
            #      partitioning.
            #   - Vertex animations refer to meshes by their names.
            ("mName", String),

            # The number of attachment meshes. Note! Currently only works with Collada loader.
            ("mNumAnimMeshes", c_uint),

            # Attachment meshes for this mesh, for vertex-based animation.
            # Attachment meshes carry replacement data for some of the
            # mesh'es vertex components (usually positions, normals).
            # Note! Currently only works with Collada loader.
            ("mAnimMeshes", POINTER(POINTER(AnimMesh))),

            # Method of morphing when animeshes are specified.
            ("mMethod", c_uint),

        ]

class Camera(Structure):
    """
    See 'camera.h' for details.
    """


    _fields_ = [
            # The name of the camera.
            #  There must be a node in the scenegraph with the same name.
            #  This node specifies the position of the camera in the scene
            #  hierarchy and can be animated.
            ("mName", String),

            # Position of the camera relative to the coordinate space
            #  defined by the corresponding node.
            #  The default value is 0|0|0.
            ("mPosition", Vector3D),

            # 'Up' - vector of the camera coordinate system relative to
            #  the coordinate space defined by the corresponding node.
            #  The 'right' vector of the camera coordinate system is
            #  the cross product of  the up and lookAt vectors.
            #  The default value is 0|1|0. The vector
            #  may be normalized, but it needn't.
            ("mUp", Vector3D),

            # 'LookAt' - vector of the camera coordinate system relative to
            #  the coordinate space defined by the corresponding node.
            #  This is the viewing direction of the user.
            #  The default value is 0|0|1. The vector
            #  may be normalized, but it needn't.
            ("mLookAt", Vector3D),

            # Half horizontal field of view angle, in radians.
            #  The field of view angle is the angle between the center
            #  line of the screen and the left or right border.
            #  The default value is 1/4PI.
            ("mHorizontalFOV", c_float),

            # Distance of the near clipping plane from the camera.
            # The value may not be 0.f (for arithmetic reasons to prevent
            # a division through zero). The default value is 0.1f.
            ("mClipPlaneNear", c_float),

            # Distance of the far clipping plane from the camera.
            # The far clipping plane must, of course, be further away than the
            # near clipping plane. The default value is 1000.f. The ratio
            # between the near and the far plane should not be too
            # large (between 1000-10000 should be ok) to avoid floating-point
            # inaccuracies which could lead to z-fighting.
            ("mClipPlaneFar", c_float),

            # Screen aspect ratio.
            # This is the ration between the width and the height of the
            # screen. Typical values are 4/3, 1/2 or 1/1. This value is
            # 0 if the aspect ratio is not defined in the source file.
            # 0 is also the default value.
            ("mAspect", c_float),
        ]

class VectorKey(Structure):
    """
    See 'anim.h' for details.
    """

    _fields_ = [
            # The time of this key
            ("mTime", c_double),

            # The value of this key
            ("mValue", Vector3D),
        ]

class QuatKey(Structure):
    """
    See 'anim.h' for details.
    """

    _fields_ = [
            # The time of this key
            ("mTime", c_double),

            # The value of this key
            ("mValue", Quaternion),
        ]

class MeshMorphKey(Structure):
    """
    See 'anim.h' for details.
    """

    _fields_ = [
            # The time of this key
            ("mTime", c_double),

            # The values and weights at the time of this key
            ("mValues", POINTER(c_uint)),
            ("mWeights", POINTER(c_double)),

            # The number of values and weights
            ("mNumValuesAndWeights", c_uint),

        ]

class NodeAnim(Structure):
    """
    See 'anim.h' for details.
    """

    _fields_ = [
            # The name of the node affected by this animation. The node
            #  must exist and it must be unique.
            ("mNodeName", String),

            # The number of position keys
            ("mNumPositionKeys", c_uint),

            # The position keys of this animation channel. Positions are
            # specified as 3D vector. The array is mNumPositionKeys in size.
            # If there are position keys, there will also be at least one
            # scaling and one rotation key.
            ("mPositionKeys", POINTER(VectorKey)),

            # The number of rotation keys
            ("mNumRotationKeys", c_uint),

            # The rotation keys of this animation channel. Rotations are
            #  given as quaternions,  which are 4D vectors. The array is
            #  mNumRotationKeys in size.
            # If there are rotation keys, there will also be at least one
            # scaling and one position key.
            ("mRotationKeys", POINTER(QuatKey)),

            # The number of scaling keys
            ("mNumScalingKeys", c_uint),

            # The scaling keys of this animation channel. Scalings are
            #  specified as 3D vector. The array is mNumScalingKeys in size.
            # If there are scaling keys, there will also be at least one
            # position and one rotation key.
            ("mScalingKeys", POINTER(VectorKey)),

            # Defines how the animation behaves before the first
            #  key is encountered.
            #  The default value is aiAnimBehaviour_DEFAULT (the original
            #  transformation matrix of the affected node is used).
            ("mPreState", c_uint),

            # Defines how the animation behaves after the last
            #  key was processed.
            #  The default value is aiAnimBehaviour_DEFAULT (the original
            #  transformation matrix of the affected node is taken).
            ("mPostState", c_uint),
        ]

class MeshAnim(Structure):
    """
    See 'anim.h' for details.
    """

    _fields_ = [
            # Name of the mesh to be animated. An empty string is not allowed,
            #  animated meshes need to be named (not necessarily uniquely,
            #  the name can basically serve as wild-card to select a group
            #  of meshes with similar animation setup)
            ("mName", String),

            # Size of the #mKeys array. Must be 1, at least.
            ("mNumKeys", c_uint),

            # Key frames of the animation. May not be NULL.
            ("mKeys", POINTER(MeshKey)),
        ]

class MeshMorphAnim(Structure):
    """
    See 'anim.h' for details.
    """

    _fields_ = [
            # Name of the mesh to be animated. An empty string is not allowed,
            # animated meshes need to be named (not necessarily uniquely,
            # the name can basically serve as wildcard to select a group
            # of meshes with similar animation setup)
            ("mName", String),

            # Size of the #mKeys array. Must be 1, at least.
            ("mNumKeys", c_uint),

            # Key frames of the animation. May not be NULL.
            ("mKeys", POINTER(MeshMorphKey)),
        ]


class Animation(Structure):
    """
    See 'anim.h' for details.
    """

    _fields_ = [
            # The name of the animation. If the modeling package this data was
            #  exported from does support only a single animation channel, this
            #  name is usually empty (length is zero).
            ("mName", String),

            # Duration of the animation in ticks.
            ("mDuration", c_double),

            # Ticks per second. 0 if not specified in the imported file
            ("mTicksPerSecond", c_double),

            # The number of bone animation channels. Each channel affects
            #  a single node.
            ("mNumChannels", c_uint),

            # The node animation channels. Each channel affects a single node.
            #  The array is mNumChannels in size.
            ("mChannels", POINTER(POINTER(NodeAnim))),

            # The number of mesh animation channels. Each channel affects
            #  a single mesh and defines vertex-based animation.
            ("mNumMeshChannels", c_uint),

            # The mesh animation channels. Each channel affects a single mesh.
            #  The array is mNumMeshChannels in size.
            ("mMeshChannels", POINTER(POINTER(MeshAnim))),

            # The number of mesh animation channels. Each channel affects
            # a single mesh and defines morphing animation.
            ("mNumMorphMeshChannels", c_uint),

            # The morph mesh animation channels. Each channel affects a single mesh.
            # The array is mNumMorphMeshChannels in size.
            ("mMorphMeshChannels", POINTER(POINTER(MeshMorphAnim))),

        ]

class ExportDataBlob(Structure):
    """
    See 'cexport.h' for details.

    Note that the '_fields_' definition is outside the class to allow the 'next' field to be recursive
    """
    pass

ExportDataBlob._fields_ = [
            # Size of the data in bytes
            ("size", c_size_t),

            # The data.
            ("data", c_void_p),

            # Name of the blob. An empty string always
            # indicates the first (and primary) blob,
            # which contains the actual file data.
            # Any other blobs are auxiliary files produced
            # by exporters (i.e. material files). Existence
            # of such files depends on the file format. Most
            # formats don't split assets across multiple files.
            #
            # If used, blob names usually contain the file
            # extension that should be used when writing
            # the data to disc.
            ("name", String),

            # Pointer to the next blob in the chain or NULL if there is none.
            ("next", POINTER(ExportDataBlob)),
        ]


class Scene(Structure):
    """
    See 'aiScene.h' for details.
    """

    AI_SCENE_FLAGS_INCOMPLETE = 0x1
    AI_SCENE_FLAGS_VALIDATED = 0x2
    AI_SCENE_FLAGS_VALIDATION_WARNING =  	0x4
    AI_SCENE_FLAGS_NON_VERBOSE_FORMAT =  	0x8
    AI_SCENE_FLAGS_TERRAIN = 0x10
    AI_SCENE_FLAGS_ALLOW_SHARED = 0x20

    _fields_ = [
            # Any combination of the AI_SCENE_FLAGS_XXX flags. By default
            # this value is 0, no flags are set. Most applications will
            # want to reject all scenes with the AI_SCENE_FLAGS_INCOMPLETE
            # bit set.
            ("mFlags", c_uint),

            # The root node of the hierarchy.
            # There will always be at least the root node if the import
            # was successful (and no special flags have been set).
            # Presence of further nodes depends on the format and content
            # of the imported file.
            ("mRootNode", POINTER(Node)),

            # The number of meshes in the scene.
            ("mNumMeshes", c_uint),

            # The array of meshes.
            # Use the indices given in the aiNode structure to access
            # this array. The array is mNumMeshes in size. If the
            # AI_SCENE_FLAGS_INCOMPLETE flag is not set there will always
            # be at least ONE material.
            ("mMeshes", POINTER(POINTER(Mesh))),

            # The number of materials in the scene.
            ("mNumMaterials", c_uint),

            # The array of materials.
            # Use the index given in each aiMesh structure to access this
            # array. The array is mNumMaterials in size. If the
            # AI_SCENE_FLAGS_INCOMPLETE flag is not set there will always
            # be at least ONE material.
            ("mMaterials", POINTER(POINTER(Material))),

            # The number of animations in the scene.
            ("mNumAnimations", c_uint),

            # The array of animations.
            # All animations imported from the given file are listed here.
            # The array is mNumAnimations in size.
            ("mAnimations", POINTER(POINTER(Animation))),

            # The number of textures embedded into the file
            ("mNumTextures", c_uint),

            # The array of embedded textures.
            # Not many file formats embed their textures into the file.
            # An example is Quake's MDL format (which is also used by
            # some GameStudio versions)
            ("mTextures", POINTER(POINTER(Texture))),

            # The number of light sources in the scene. Light sources
            # are fully optional, in most cases this attribute will be 0
            ("mNumLights", c_uint),

            # The array of light sources.
            # All light sources imported from the given file are
            # listed here. The array is mNumLights in size.
            ("mLights", POINTER(POINTER(Light))),

            # The number of cameras in the scene. Cameras
            # are fully optional, in most cases this attribute will be 0
            ("mNumCameras", c_uint),

            # The array of cameras.
            # All cameras imported from the given file are listed here.
            # The array is mNumCameras in size. The first camera in the
            # array (if existing) is the default camera view into
            # the scene.
            ("mCameras", POINTER(POINTER(Camera))),

            # This data contains global metadata which belongs to the scene like
            # unit-conversions, versions, vendors or other model-specific data. This
            # can be used to store format-specific metadata as well.
            ("mMetadata", POINTER(Metadata)),

            # Internal data, do not touch
            ("mPrivate", c_char_p),
        ]

assimp_structs_as_tuple = (Matrix4x4,
                           Matrix3x3,
                           Vector2D,
                           Vector3D,
                           Color3D,
                           Color4D,
                           Quaternion,
                           Plane,
                           Texel)



"""
PyAssimp

This is the main-module of PyAssimp.
"""

import sys
if sys.version_info < (3, 0):
    raise RuntimeError('PipAssimp needs python3 or higher')

import numpy
import logging
import ctypes
logger = logging.getLogger("pyassimp")
# attach default null handler to logger so it doesn't complain
# even if you don't attach another handler to logger
logger.addHandler(logging.NullHandler())

from . import structs
from . import helper
from . import postprocess
from .errors import AssimpError

class AssimpLib:
    """
    Assimp-Singleton
    """
    load, load_mem, export, export_blob, release, dll = helper.search_library()
_assimp_lib = AssimpLib()

def make_tuple(ai_obj, type = None):
    res = None

    #notes:
    # ai_obj._fields_ = [ ("attr", c_type), ... ]
    # getattr(ai_obj, e[0]).__class__ == float

    if isinstance(ai_obj, structs.Matrix4x4):
        if numpy:
            res = numpy.array([getattr(ai_obj, e[0]) for e in ai_obj._fields_]).reshape((4,4))
            #import pdb;pdb.set_trace()
        else:
            res = [getattr(ai_obj, e[0]) for e in ai_obj._fields_]
            res = [res[i:i+4] for i in range(0, 16, 4)]
    elif isinstance(ai_obj, structs.Matrix3x3):
        if numpy:
            res = numpy.array([getattr(ai_obj, e[0]) for e in ai_obj._fields_]).reshape((3,3))
        else:
            res = [getattr(ai_obj, e[0]) for e in ai_obj._fields_]
            res = [res[i:i+3] for i in range(0,9,3)]
    else:
        if numpy:
            res = numpy.array([getattr(ai_obj, e[0]) for e in ai_obj._fields_])
        else:
            res = [getattr(ai_obj, e[0]) for e in ai_obj._fields_]

    return res

# Returns unicode object for Python 2, and str object for Python 3.
def _convert_assimp_string(assimp_string):
    if sys.version_info < (3, 0):
        raise AssimpError('PipAssimp requires python3')
    return str(assimp_string.data, errors='ignore')

# It is faster and more correct to have an init function for each assimp class
def _init_face(aiFace):
    aiFace.indices = [aiFace.mIndices[i] for i in range(aiFace.mNumIndices)]
assimp_struct_inits =  { structs.Face : _init_face }

def call_init(obj, caller = None):
    if helper.hasattr_silent(obj,'contents'): #pointer
        _init(obj.contents, obj, caller)
    else:
        _init(obj,parent=caller)

def _is_init_type(obj):

    if obj and helper.hasattr_silent(obj,'contents'): #pointer
        return _is_init_type(obj[0])
    # null-pointer case that arises when we reach a mesh attribute
    # like mBitangents which use mNumVertices rather than mNumBitangents
    # so it breaks the 'is iterable' check.
    # Basically:
    # FIXME!
    elif not bool(obj):
        return False
    tname = obj.__class__.__name__
    return not (tname[:2] == 'c_' or tname == 'Structure' \
            or tname == 'POINTER') and not isinstance(obj, (int, str, bytes))

def _init(self, target = None, parent = None):
    """
    Custom initialize() for C structs, adds safely accessible member functionality.

    :param target: set the object which receive the added methods. Useful when manipulating
    pointers, to skip the intermediate 'contents' deferencing.
    """
    if not target:
        target = self

    dirself = dir(self)
    for m in dirself:

        if m.startswith("_"):
            continue

        if m.startswith('mNum'):
            if 'm' + m[4:] in dirself:
                continue # will be processed later on
            else:
                name = m[1:].lower()

                obj = getattr(self, m)
                setattr(target, name, obj)
                continue

        if m == 'mName':
            target.name = str(_convert_assimp_string(self.mName))
            target.__class__.__repr__ = lambda x: str(x.__class__) + "(" + getattr(x, 'name','') + ")"
            target.__class__.__str__ = lambda x: getattr(x, 'name', '')
            continue

        name = m[1:].lower()

        obj = getattr(self, m)

        # Create tuples
        if isinstance(obj, structs.assimp_structs_as_tuple):
            setattr(target, name, make_tuple(obj))
            logger.debug(str(self) + ": Added array " + str(getattr(target, name)) +  " as self." + name.lower())
            continue

        if m.startswith('m'):

            if name == "parent":
                setattr(target, name, parent)
                logger.debug("Added a parent as self." + name)
                continue

            if helper.hasattr_silent(self, 'mNum' + m[1:]):

                length =  getattr(self, 'mNum' + m[1:])

                # -> special case: properties are
                # stored as a dict.
                if m == 'mProperties':
                    setattr(target, name, _get_properties(obj, length))
                    continue


                if not length: # empty!
                    setattr(target, name, [])
                    logger.debug(str(self) + ": " + name + " is an empty list.")
                    continue


                try:
                    if obj._type_ in structs.assimp_structs_as_tuple:
                        if numpy:
                            setattr(target, name, numpy.array([make_tuple(obj[i]) for i in range(length)], dtype=numpy.float32))

                            logger.debug(str(self) + ": Added an array of numpy arrays (type "+ str(type(obj)) + ") as self." + name)
                        else:
                            setattr(target, name, [make_tuple(obj[i]) for i in range(length)])

                            logger.debug(str(self) + ": Added a list of lists (type "+ str(type(obj)) + ") as self." + name)

                    else:
                        setattr(target, name, [obj[i] for i in range(length)]) #TODO: maybe not necessary to recreate an array?

                        logger.debug(str(self) + ": Added list of " + str(obj) + " " + name + " as self." + name + " (type: " + str(type(obj)) + ")")

                        # initialize array elements
                        try:
                            init = assimp_struct_inits[type(obj[0])]
                        except KeyError:
                            if _is_init_type(obj[0]):
                                for e in getattr(target, name):
                                    call_init(e, target)
                        else:
                            for e in getattr(target, name):
                                init(e)


                except IndexError:
                    logger.error("in " + str(self) +" : mismatch between mNum" + name + " and the actual amount of data in m" + name + ". This may be due to version mismatch between libassimp and pyassimp. Quitting now.")
                    sys.exit(1)

                except ValueError as e:

                    logger.error("In " + str(self) +  "->" + name + ": " + str(e) + ". Quitting now.")
                    if "setting an array element with a sequence" in str(e):
                        logger.error("Note that pyassimp does not currently "
                                     "support meshes with mixed triangles "
                                     "and quads. Try to load your mesh with"
                                     " a post-processing to triangulate your"
                                     " faces.")
                    raise e



            else: # starts with 'm' but not iterable
                setattr(target, name, obj)
                logger.debug("Added " + name + " as self." + name + " (type: " + str(type(obj)) + ")")

                if _is_init_type(obj):
                    call_init(obj, target)

    if isinstance(self, structs.Mesh):
        _finalize_mesh(self, target)

    if isinstance(self, structs.Texture):
        _finalize_texture(self, target)

    if isinstance(self, structs.Metadata):
        _finalize_metadata(self, target)


    return self


def pythonize_assimp(type, obj, scene):
    """ This method modify the Assimp data structures
    to make them easier to work with in Python.

    Supported operations:
     - MESH: replace a list of mesh IDs by reference to these meshes
     - ADDTRANSFORMATION: add a reference to an object's transformation taken from their associated node.

    :param type: the type of modification to operate (cf above)
    :param obj: the input object to modify
    :param scene: a reference to the whole scene
    """

    if type == "MESH":
        meshes = []
        for i in obj:
            meshes.append(scene.meshes[i])
        return meshes

    if type == "ADDTRANSFORMATION":
        def getnode(node, name):
            if node.name == name: return node
            for child in node.children:
                n = getnode(child, name)
                if n: return n

        node = getnode(scene.rootnode, obj.name)
        if not node:
            raise AssimpError("Object " + str(obj) + " has no associated node!")
        setattr(obj, "transformation", node.transformation)

def recur_pythonize(node, scene):
    '''
    Recursively call pythonize_assimp on
    nodes tree to apply several post-processing to
    pythonize the assimp datastructures.
    '''
    node.meshes = pythonize_assimp("MESH", node.meshes, scene)
    for mesh in node.meshes:
        mesh.material = scene.materials[mesh.materialindex]
    for cam in scene.cameras:
        pythonize_assimp("ADDTRANSFORMATION", cam, scene)
    for c in node.children:
        recur_pythonize(c, scene)

def load(filename,
         file_type = None,
         processing = postprocess.aiProcess_Triangulate):
    '''
    Load a model into a scene. On failure throws AssimpError.

    Arguments
    ---------
    filename:   Either a filename or a file object to load model from.
                If a file object is passed, file_type MUST be specified
                Otherwise Assimp has no idea which importer to use.
                This is named 'filename' so as to not break legacy code.
    processing: assimp postprocessing parameters. Verbose keywords are imported
                from postprocessing, and the parameters can be combined bitwise to
                generate the final processing value. Note that the default value will
                triangulate quad faces. Example of generating other possible values:
                processing = (pyassimp.postprocess.aiProcess_Triangulate |
                              pyassimp.postprocess.aiProcess_OptimizeMeshes)
    file_type:  string of file extension, such as 'stl'

    Returns
    ---------
    Scene object with model data
    '''

    if hasattr(filename, 'read'):
        # This is the case where a file object has been passed to load.
        # It is calling the following function:
        # const aiScene* aiImportFileFromMemory(const char* pBuffer,
        #                                      unsigned int pLength,
        #                                      unsigned int pFlags,
        #                                      const char* pHint)
        if file_type is None:
            raise AssimpError('File type must be specified when passing file objects!')
        data = filename.read()
        model = _assimp_lib.load_mem(data,
                                     len(data),
                                     processing,
                                     file_type)
    else:
        # a filename string has been passed
        model = _assimp_lib.load(filename.encode(sys.getfilesystemencoding()), processing)

    if not model:
        raise AssimpError('Could not import file!')
    scene = _init(model.contents)
    recur_pythonize(scene.rootnode, scene)
    return scene

def export(scene,
           filename,
           file_type = None,
           processing = postprocess.aiProcess_Triangulate):
    '''
    Export a scene. On failure throws AssimpError.

    Arguments
    ---------
    scene: scene to export.
    filename: Filename that the scene should be exported to.
    file_type: string of file exporter to use. For example "collada".
    processing: assimp postprocessing parameters. Verbose keywords are imported
                from postprocessing, and the parameters can be combined bitwise to
                generate the final processing value. Note that the default value will
                triangulate quad faces. Example of generating other possible values:
                processing = (pyassimp.postprocess.aiProcess_Triangulate |
                              pyassimp.postprocess.aiProcess_OptimizeMeshes)

    '''

    exportStatus = _assimp_lib.export(ctypes.pointer(scene), file_type.encode("ascii"), filename.encode(sys.getfilesystemencoding()), processing)

    if exportStatus != 0:
        raise AssimpError('Could not export scene!')

def export_blob(scene,
                file_type=None,
                processing=postprocess.aiProcess_Triangulate):
    '''
    Export a scene and return a blob in the correct format. On failure throws AssimpError.

    Arguments
    ---------
    scene: scene to export.
    file_type: string of file exporter to use. For example "collada".
    processing: assimp postprocessing parameters. Verbose keywords are imported
                from postprocessing, and the parameters can be combined bitwise to
                generate the final processing value. Note that the default value will
                triangulate quad faces. Example of generating other possible values:
                processing = (pyassimp.postprocess.aiProcess_Triangulate |
                              pyassimp.postprocess.aiProcess_OptimizeMeshes)
    Returns
    ---------
    Pointer to structs.ExportDataBlob
    '''
    exportBlobPtr = _assimp_lib.export_blob(ctypes.pointer(scene), file_type.encode("ascii"), processing)

    if exportBlobPtr == 0:
        raise AssimpError('Could not export scene to blob!')
    return exportBlobPtr

def release(scene):
    _assimp_lib.release(ctypes.pointer(scene))

def _finalize_texture(tex, target):
    setattr(target, "achformathint", tex.achFormatHint)
    if numpy:
        data = numpy.array([make_tuple(getattr(tex, "pcData")[i]) for i in range(tex.mWidth * tex.mHeight)])
    else:
        data = [make_tuple(getattr(tex, "pcData")[i]) for i in range(tex.mWidth * tex.mHeight)]
    setattr(target, "data", data)

def _finalize_mesh(mesh, target):
    """ Building of meshes is a bit specific.

    We override here the various datasets that can
    not be process as regular fields.

    For instance, the length of the normals array is
    mNumVertices (no mNumNormals is available)
    """
    nb_vertices = getattr(mesh, "mNumVertices")

    def fill(name):
        mAttr = getattr(mesh, name)
        if numpy:
            if mAttr:
                data = numpy.array([make_tuple(getattr(mesh, name)[i]) for i in range(nb_vertices)], dtype=numpy.float32)
                setattr(target, name[1:].lower(), data)
            else:
                setattr(target, name[1:].lower(), numpy.array([], dtype="float32"))
        else:
            if mAttr:
                data = [make_tuple(getattr(mesh, name)[i]) for i in range(nb_vertices)]
                setattr(target, name[1:].lower(), data)
            else:
                setattr(target, name[1:].lower(), [])

    def fillarray(name):
        mAttr = getattr(mesh, name)

        data = []
        for index, mSubAttr in enumerate(mAttr):
            if mSubAttr:
                data.append([make_tuple(getattr(mesh, name)[index][i]) for i in range(nb_vertices)])

        if numpy:
            setattr(target, name[1:].lower(), numpy.array(data, dtype=numpy.float32))
        else:
            setattr(target, name[1:].lower(), data)

    fill("mNormals")
    fill("mTangents")
    fill("mBitangents")

    fillarray("mColors")
    fillarray("mTextureCoords")

    # prepare faces
    if numpy:
        faces = numpy.array([f.indices for f in target.faces], dtype=numpy.int32)
    else:
        faces = [f.indices for f in target.faces]
    setattr(target, 'faces', faces)

def _init_metadata_entry(entry):
    entry.type = entry.mType
    if entry.type == structs.MetadataEntry.AI_BOOL:
        entry.data = ctypes.cast(entry.mData, ctypes.POINTER(ctypes.c_bool)).contents.value
    elif entry.type == structs.MetadataEntry.AI_INT32:
        entry.data = ctypes.cast(entry.mData, ctypes.POINTER(ctypes.c_int32)).contents.value
    elif entry.type == structs.MetadataEntry.AI_UINT64:
        entry.data = ctypes.cast(entry.mData, ctypes.POINTER(ctypes.c_uint64)).contents.value
    elif entry.type == structs.MetadataEntry.AI_FLOAT:
        entry.data = ctypes.cast(entry.mData, ctypes.POINTER(ctypes.c_float)).contents.value
    elif entry.type == structs.MetadataEntry.AI_DOUBLE:
        entry.data = ctypes.cast(entry.mData, ctypes.POINTER(ctypes.c_double)).contents.value
    elif entry.type == structs.MetadataEntry.AI_AISTRING:
        assimp_string = ctypes.cast(entry.mData, ctypes.POINTER(structs.String)).contents
        entry.data = _convert_assimp_string(assimp_string)
    elif entry.type == structs.MetadataEntry.AI_AIVECTOR3D:
        assimp_vector = ctypes.cast(entry.mData, ctypes.POINTER(structs.Vector3D)).contents
        entry.data = make_tuple(assimp_vector)

    return entry

def _finalize_metadata(metadata, target):
    """ Building the metadata object is a bit specific.

    Firstly, there are two separate arrays: one with metadata keys and one
    with metadata values, and there are no corresponding mNum* attributes,
    so the C arrays are not converted to Python arrays using the generic
    code in the _init function.

    Secondly, a metadata entry value has to be cast according to declared
    metadata entry type.
    """
    length = metadata.mNumProperties
    setattr(target, 'keys', [str(_convert_assimp_string(metadata.mKeys[i])) for i in range(length)])
    setattr(target, 'values', [_init_metadata_entry(metadata.mValues[i]) for i in range(length)])

class PropertyGetter(dict):
    def __getitem__(self, key):
        semantic = 0
        if isinstance(key, tuple):
            key, semantic = key

        return dict.__getitem__(self, (key, semantic))

    def keys(self):
        for k in dict.keys(self):
            yield k[0]

    def __iter__(self):
        return self.keys()

    def items(self):
        for k, v in dict.items(self):
            yield k[0], v


def _get_properties(properties, length):
    """
    Convenience Function to get the material properties as a dict
    and values in a python format.
    """
    result = {}
    #read all properties
    for p in [properties[i] for i in range(length)]:
        #the name
        p = p.contents
        key = str(_convert_assimp_string(p.mKey))
        key = (key.split('.')[1], p.mSemantic)

        #the data
        if p.mType == 1:
            arr = ctypes.cast(p.mData,
                              ctypes.POINTER(ctypes.c_float * int(p.mDataLength/ctypes.sizeof(ctypes.c_float)))
                              ).contents
            value = [x for x in arr]
        elif p.mType == 3: #string can't be an array
            value = _convert_assimp_string(ctypes.cast(p.mData, ctypes.POINTER(structs.MaterialPropertyString)).contents)

        elif p.mType == 4:
            arr = ctypes.cast(p.mData,
                              ctypes.POINTER(ctypes.c_int * int(p.mDataLength/ctypes.sizeof(ctypes.c_int)))
                              ).contents
            value = [x for x in arr]
        else:
            value = p.mData[:p.mDataLength]

        if len(value) == 1:
            [value] = value

        result[key] = value

    return PropertyGetter(result)

def decompose_matrix(matrix):
    if not isinstance(matrix, structs.Matrix4x4):
        raise AssimpError("pyassimp.decompose_matrix failed: Not a Matrix4x4!")

    scaling = structs.Vector3D()
    rotation = structs.Quaternion()
    position = structs.Vector3D()

    _assimp_lib.dll.aiDecomposeMatrix(ctypes.pointer(matrix),
                                      ctypes.byref(scaling),
                                      ctypes.byref(rotation),
                                      ctypes.byref(position))
    return scaling._init(), rotation._init(), position._init()




from setuptools import setup, find_packages
setup(
    name = 'PipAssimp',         # How you named your package folder (MyLib)
    packages = ['PipAssimp'],   # Chose the same as "name"
    version = '1.0.5',      # Start with a small number and increase it with every change you make
    license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description = 'Conveniece pkg for install assimp/pyassimp from requirements.txt',   # Give a short description about your library
    author = 'James Ford',                   # Type in your name
    author_email = 'jameslford8@gmail.com',      # Type in your E-Mail
    url = 'https://github.com/jameslford/PipAssimp',   # Provide either the link to your github or to your website
    download_url = 'https://github.com/jameslford/PipAssimp/archive/v0.0.1.tar.gz',    # I explain this later on
    keywords = ['3D', 'File Conversion', 'gltf', 'glb', 'obj'],   # Keywords that define your package best
    install_requires=['numpy'],
    # # package_data = {'': ['libassimp.so.5.0.1']},
    # data_files = [('', 'libassimp.so.5.0.1')],
    # include_package_data=True,
    classifiers=[
      'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
      'Intended Audience :: Developers',      # Define that your audience are developers
      'Topic :: Software Development :: Build Tools',
      'License :: OSI Approved :: MIT License',   # Again, pick a license
      'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
      'Programming Language :: Python :: 3.4',
      'Programming Language :: Python :: 3.5',
      'Programming Language :: Python :: 3.6',
  ],
)